## Tree
### 一个无限层级的树结构插件，支持数据的属性自定义，任意层级的嵌套，节点数量无限多，并且处理大量数据的时候无卡顿。

### 安装

    npm install tree

### 使用

    // ES6
    import tree from 'tree.js'
    // require
    var tree = require('tree')

    Vue.use(tree)

    // 或者直接使用script导入
    <script src="./node_modules/vue/dist/my-tree.js"></script>

    // 作为组件的方式使用
    <tree></tree>
### 配置

    <tree
        idKey="id"
        contentKey="content"
        parentKey="parentId"
        :data="treeData"
        :rootId="rootId"
        :loadData="loadData"
        @itemClick="itemClick"
    />

    export default {
        data(){
        return {
                treeData: [
                    { id: 1, content: '1', parentId: null },
                    { id: 2, content: '1-1', parentId: 1 },
                    { id: 3, content: '1-2', parentId: 1 },
                    { id: 4, content: '1-2-1', parentId: 3 },
                    { id: 5, content: '1-2-2', parentId: 3 },
                    { id: 6, content: '1-3', parentId: 1 }
                ],
                rootId: 1
            }
        },
        methods: {
            itemClick(item){
                console.log(item.content)
            },
        loadData(item) {
                return new Promise((resolve, reject) => {
                    setTimeout(() => {
                    if (item.id === 6) {
                        // has data
                        resolve([{ id: 7, content: '1-3-1', parentId: 6 }])
                    } else if (item.id === 5) {
                        // assume there's a network error
                        reject()
                    } else {
                        // doesn't have data
                        resolve([])
                    }
                    }, 3000)
                })
            }

        },
        created(){
            let list = utils.createTreeNode(10, 10000, this.rootId)
            console.log(list)
            this.treeData = list; 
            //模拟更改变换数据
            setTimeout(()=> {
                console.log('更改数据')
                this.rootId = 2;
                this.treeData = utils.createTreeNode(10, 10000, this.rootId)
            }, 3000)
        }

    }

### Props
参数|说明|是否必填|类型|默认值
-|-|-|-|-
idKey|数据中每个元素的唯一标识|N|string|'id'
contentKey|数据的内容标识|N|stirng|'content'
parentKey|数据的父元素标识|N|string|'parent'
data|原始数据|Y|any[]|--
pageNum|每页展示的节点数|N|number|20
rootId|做为根的数据标识|N|number|--
itemClick|当点击某一节点时触发的事件|N|(item:any) => any|null
loadData|懒加载所需的函数，若非null则树需要支持懒加载|N|(item:any) => Promise <any[]\|null>|null

### 建树的算法

思路：
1. 利用两层for循环遍历查找每个节点的children（废弃）

代码如下：

     /**
        * 将扁平数据转为树级结构的数据，时间复杂度为o(n^2)，数据量大的时候，生成树级结构需要比较长的时间
        * 思路：用两重for循环遍历数据，先找到每个父节点的字节点数组，双重遍历，返回根节点
        * 抛弃这算法，因为数据量大的时候，明显耗时很多
        */
    // setTreeData(data, rootId){
    //     //克隆原数据
    //     let cloneData = JSON.parse(JSON.stringify(data));
    //     return cloneData.filter((father) => {
    //         //默认isShow为true
    //         if(father.isShow === undefined){
    //             father.isShow = true
    //         }
    //         //遍历所有数据，生成children
    //         let branchArr = cloneData.filter((children) => {
    //             //非前两层，则isShow为false
    //             if(father[this.idKey] == children[this.parentKey] && father[this.parentKey] && this.rootId){
    //                 children.isShow = false;
    //             }
    //             //返回每一项的子级数组
    //             return father[this.idKey] == children[this.parentKey]
    //         });
    //         //给父级数据添加children属性，并赋值
    //         father.children = branchArr.length > 0 ? branchArr : [];
    //         // branchArr.length > 0 ? father.children = branchArr : [];
    //         return father[this.idKey] == rootId;
    //     })
    // }

缺点：时间复杂度高，O(n^2)，当数据量大的时候，处理时间长，页面卡顿严重。
2. 由于原始数据中id唯一，并且有parentId指向父节点，可先将原始数据以id为key，map为一个set结构，对set结构进行遍历，并提供parentId查找到父节点，将当前节点的数据存入父节点的children中。

代码如下：

     /**
        * 改进之后的算法，时间复杂度O(n)
        * 思路: 遍历所有数据，由于id为唯一，所以生成由id为key的set结构
        * 遍历set结构，由于每一个数据都有一个parentId，可由set[data.parentId]获取到当前数据的父节点
        * 将当前数据存入到父节点的children中
        */
    setTreeData(data, rootId){
        //先生成set结构的数据
        let result = {};
        let cloneData = JSON.parse(JSON.stringify(data));
        cloneData.filter(data => {
            //id是唯一的，
            //给对象添加children属性
            data.children = [];
            data.isShow = true;
            result[data[this.idKey]] = data;
        });
        //遍历result对象
        for(let item in result){
            let temp = result[item]
            //将当前对象存入父节点的children
            if(temp[this.parentKey]){
                //不是前两层则默认隐藏
                if(temp[this.parentKey]!= rootId) temp.isShow = false;
                let parent = result[temp[this.parentKey]]
                if(parent){
                    let childLength =  parent.children.length;
                    parent.children[childLength] = temp;
                }
            }
        }
        return [result[rootId]];
    }
优点：只用了一层循环遍历（2次）算法的时间复杂度是O(n)。当数据量较大的时候，处理时间比第一种方法快很多，页面没有明显卡顿，效率提高不少。代码实现简单，思路清晰可理解。

## 虚拟DOM
页面的DOM是采用虚拟DOM实现，利用render函数实现虚拟DOM，比真实DOM结构性能高。因为每次数据变动，虚拟DOM只会更新变动的一小部分。

## 懒加载
配置loadData，实现懒加载。
在点击当前节点的时候，即时加载当前节点的字节点数据。

## 优化无限层级滚动效率
利用Intersection Observer优化滚动效率

思路：控制当前页面渲染的DOM节点。只有在滚动条滚动到页面底部时，才会加载剩余节点的一部分。本组件中默认是20个节点为一组。

原理：利用Intersection Observer，观察当前页面的节点，当节点与视图（viewport）交叉，并且可见时，触发回调。在回调函数中，加载剩余的数据。
